#include <sys/io.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define ALSA_INFO_FILE "/proc/asound/card%i/%s"

#define MAX_N_CARD 16
#define MAX_CARD_SCAN_LEN 32

int readCardPort (int nCard, char *cardId) {
  char cardInfoName[255];
  sprintf (cardInfoName, ALSA_INFO_FILE, nCard, cardId);
  FILE *cardInfoFile = fopen (cardInfoName, "r");

  int port = -1;

  if (cardInfoFile != NULL) {
      fprintf (stderr, "Found %s ...\n", cardInfoName);
      int scanned = 0;
      int found;
      char value[33];
      while ((found =
              fscanf (cardInfoFile, "%31s at 0x%x", (char *) &value,
                      &port)) > 0) {
          scanned += found;
          if (found > 1 || scanned >= MAX_CARD_SCAN_LEN) {
              break;
          }
      }
      fprintf (stderr, "%s at 0x%x.\n", cardId, port);

      fclose (cardInfoFile);
  }
  return port;
}

#define ALSA_CHANNEL_FILE "/proc/asound/card%i/pcm%ip/sub0/hw_params"

FILE *openAlsaChannelFile (int nCard, int nChannel) {
  char cardName[255];
  sprintf (cardName, ALSA_CHANNEL_FILE, nCard, nChannel);
  fprintf (stderr, "Checking %s ...\n", cardName);
  FILE *cardFile = fopen (cardName, "r");

  return cardFile;
}

int readAlsaChannelRate (FILE * cardFile) {
  char name[33];
  int scanned = 0;
  int n;
  int rate = 0;

  if(fseek(cardFile, 0, SEEK_SET)) {
      printf("Seek error.\n");
      return -1;
  }
  fflush(cardFile);

  while ((n = fscanf (cardFile, "%31s%i\n", (char *) &name, &rate)) > 0) {
      if (scanned == 0 && !strncmp (name, "closed", 32)) {
          return 0;
      }
      scanned += n;
//      fprintf(stderr, "--> %i '%s' %i\n", n, name, rate);
      if (n == 2 && !strncmp (name, "rate:", 32)) {
          break;
      }
  }
  return rate;
}

#define GPIO_DATA_LOWER  0x14
#define GPIO_DATA_HIGHER 0x15
#define GPIO_READ_LOWER  0x16
#define GPIO_READ_HIGHER 0x17
#define GPIO_DIR_LOWER   0x18
#define GPIO_DIR_HIGHER  0x19

#define GPIO_DATA GPIO_DATA_HIGHER
#define GPIO_READ GPIO_READ_HIGHER
#define GPIO_DIR  GPIO_DIR_HIGHER


int askPortPermission(int basePort) {
    return ioperm(basePort+0x04, 1, 1) // CCS04
        || ioperm(basePort+0x05, 1, 1) // CCS05
        || ioperm(basePort+0x06, 1, 1) // CCS06
        || ioperm(basePort+0x07, 1, 1) // CCS07
        || ioperm(basePort+GPIO_DATA,1,1)
        || ioperm(basePort+GPIO_READ,1,1)
        || ioperm(basePort+GPIO_DIR,1,1);
}

#define BIT_CD     0x1
#define BIT_HIRES  0x2
#define BIT_192    0x4
#define BIT_BLINK  0x8

#define BIT_MASK   BIT_CD | BIT_HIRES | BIT_192 | BIT_BLINK

void initPort(int basePort) {
    unsigned char ccs04 = inb(basePort + 0x04);
    unsigned char ccs05 = inb(basePort + 0x05);
    unsigned char ccs06 = inb(basePort + 0x06);
    unsigned char ccs07 = inb(basePort + 0x07);

    fprintf(stderr, "CCS04-CCS07 bytes are %02x %02x %02x %02x.\n",
            ccs04, ccs05, ccs06, ccs07);
    fprintf(stderr, "Using GPIO port 0x%x.\n", basePort + GPIO_DATA);
    unsigned char dir = inb(basePort+GPIO_DIR);
    fprintf(stderr, "Direction 0x%02x (were 0x%02x).\n", dir | BIT_MASK, dir);
    outb(dir | BIT_MASK, basePort+GPIO_DIR);
}


void writePort(int basePort, unsigned char value) {
    outb(value, basePort+GPIO_DATA);
}

#define CD_FREQUENCY 44100
#define DVD_FREQUENCY 48000

unsigned char rateToBits(int rate) {
    unsigned char cdBase = 0;
    int multiplier = 1;
    if (rate % CD_FREQUENCY == 0) {
        multiplier = rate / CD_FREQUENCY;
        cdBase = 1;
        fprintf(stderr, "Using CD-based x%i frequency %i.\n",multiplier, rate);
    } else if (rate % DVD_FREQUENCY == 0) {
        multiplier = rate / DVD_FREQUENCY;
        fprintf(stderr, "Using DVD-based x%i frequency %i.\n",multiplier, rate);
    } else {
        multiplier = rate >= DVD_FREQUENCY ? rate / DVD_FREQUENCY : 1;
        fprintf(stderr, "Using UNKNOWN x%i frequency %i.\n",multiplier, rate);
    }
    unsigned char bits = cdBase ? BIT_CD : 0;
    if (multiplier == 2) {
        bits |= BIT_HIRES;
    } else if (multiplier == 4) {
        bits |= BIT_HIRES;
        bits |= BIT_192;
    } else if (multiplier > 1) {
        bits |= BIT_HIRES;
        fprintf(stderr, "Strange multiplier x%i!\n", multiplier);
    }
    return bits;
}

int main (int argc, char *argv[]) {
  char *cardId = "ice1724";
  int nChannel = 0;
  int nCard = -1;
  if (argc > 4 || (argc == 2 && (!strncmp (argv[1], "-h", 32)
                                 || !strncmp (argv[1], "--help", 32)))) {
        printf("Usage:\n%s [channel] [chip_id] [card] \n", argv[0]);
        printf("Defaults:\n%s %i %s %i\n", argv[0], nChannel, cardId, nCard);
        printf("(card -1 means autoscan)\n");
        return 0;
  }
  if(argc > 1 && strlen(argv[1]) > 0){
       nChannel = atoi(argv[1]);
       fprintf(stderr, "Set channel %i\n", nChannel);
  }
  if(argc > 2 && strlen(argv[2]) > 0){
       cardId = argv[2];
       fprintf(stderr, "Set chip_idl '%s'\n", cardId);
  }
  if(argc > 3 && strlen(argv[3]) > 0){
       nCard = atoi(argv[3]);
       fprintf(stderr, "Set card %i\n", nCard);
  }

  int usePort = -1;
  int n;
  for (n= 0; n < MAX_N_CARD; n++) {

      int port = readCardPort (n, cardId);
      if (port != -1 && (nCard == -1 || nCard == n)) {
          nCard = n;
          usePort = port;
          fprintf (stderr, "Card used as %i at port 0x%x.\n", nCard,
                   usePort);
          break;
      }
  }

  FILE *cardFile = openAlsaChannelFile (nCard, nChannel);

  if (cardFile == NULL) {
      fprintf (stderr, "Cannot open alsa info file.\n");
      return 1;
  }


  if (askPortPermission(usePort) != 0) {
      fprintf(stderr, "Cannot get permissions to ports at 0x%x.\n", usePort);
      fprintf(stderr, "Program must be run with root privilegies.\n");
      return 1;
  }
  initPort(usePort);

  int rate;

  struct timespec delay;
  delay.tv_sec = 0;
  delay.tv_nsec = 100000000;
  int previousRate = -1; // never happens yet
  unsigned char portValue = 0; // default
  unsigned char blink = 0;
  while (1) {
      rate = readAlsaChannelRate (cardFile);
      if (-1 == rate) { // seek error
          return 1;
      }
      if (rate != previousRate) {
          if (rate == 0) {
              fprintf (stderr, "Stopping...\n");
          } else {
              portValue = rateToBits(rate);
          }
          previousRate = rate;
      }
      blink ^= 1;
      unsigned char writeToPort = blink ? portValue | BIT_BLINK : portValue;;
      writePort(usePort, writeToPort);
      nanosleep (&delay, NULL);
  }
  fclose (cardFile);
  return 0;
}
